/* Serial line terminal program
 * characters can be displayed as in hex, decimal and ascii 
 * characters can be seperated by space, tab, newline or 'nothing'
 * local echo can be switched on or off
 *

 * by Albrecht Schmidt, Lancaster University - Oct 2001
 * http://www.comp.lancs.ac.uk/~albrecht 
 * Albrecht@comp.lancs.ac.uk
 * based on an example from Robert Mashlan - 
 *                    see http://r2m.com/~rmashlan/
 */

#define STRICT
#include "robot_control.h"
#include "serialterm.h"
#include "stdafx.h"
#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

extern var_robot *variables;
// global vars - to aviod parameter passing over thread
// I am lazy I know ...
char exename[1024];
char logname[1024];
int DisplayMode, separator, echo, logo;

// global vars
char datos_brutos[100];
char ident[]="00000";
char ident1[]="GPGGA";
char latitud[]="00000000000";
char latitud_dir;
char longitud[]="000000000000";
char longitud_dir;
char altitud[]="0000000";
char hora[]="000000000";
int comas[14];
int cont1=0;
int aux1=0;
int flag1=0;
char hor[]="00";
char min[]="00";
char seg[]="00";
char lat_hor[]="00";
char lat_min_ent[]="00";
char lat_min_dec[]="000000";
char long_hor[]="00";
char long_min_ent[]="00";
char long_min_dec[]="000000";
char bufo[1700];
int j,k,i1;
int cont=0;
char c;
int lat_deg,long_deg,alt,hora1,hora2,hora3;
int a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,c1,c2,c3,d1,d2,e1,e2,f1,f2;
long int radio=6378140;
float x[2];
float y[2];
double desp, desp_x=0, desp_y=0, aux2, Xin, Yin, Xout, Yout,A,B,delta;
double PHI,THETA,latmin,longmin,lambda,eps,nu,ve,psi,AA1,AA2,J2,J4,J6,alpha,beta,gamma,BePHI,xUTM,yUTM,Xmed,Ymed;
double er2=0.0067681703;
int long_central=-69;
long int const_a=6378388;
long int const_c=6399936.609;
int contador=0;
extern char *puerto;

void help() {
  printf("%s port [speed] [DisplayMode] [Separator] [Echo] [logfilename]\n", exename);
  printf("        port ::= com1 | com2 | com3 | com4 | com5 | com6\n");
  printf("        speed::= 300 | 4800 | 9600 | 19200 | 38400 | 57600 | 115200 | 230400\n");
  printf("                                     ^^^^^\n");
  printf("        DisplayMode::= ascii | hex | decimal\n");
  printf("                       ^^^^^\n");
  printf("        Separator::= empty | space | newline | tab\n");
  printf("                     ^^^^^\n");
  printf("        Echo::= no | yes\n");
  printf("                ^^\n");
  printf("        logfilename::= <anyname> (if not provided no log is written)\n\n");
  printf("Example: %s com1 115200 hex space no log.txt\n", exename);
  printf("    open the terminal on port com1 with 115200 bit/s, print hex code of\n"); 
  printf("    incoming characters, seperate them by space, no local echo, log to log.txt\n");
  printf("Example: %s com2 19200 decimal tab yes\n", exename);
  printf("    open the terminal on port com2 with 19200 bit/s, print decimal code\n"); 
  printf("    of incoming characters, seperate them by tabs, do local echo, no logfile\n");
}


void PrintError( LPCSTR str)
{
   LPVOID lpMessageBuffer;
   int error = GetLastError();
   FormatMessage(
      FORMAT_MESSAGE_ALLOCATE_BUFFER |
      FORMAT_MESSAGE_FROM_SYSTEM,
      NULL,
      error,
      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), //The user default language
      (LPTSTR) &lpMessageBuffer,
      0,
      NULL
   );
   printf("%s: (%d) %s\n\n",str,error,lpMessageBuffer);
   help();
   LocalFree( lpMessageBuffer );
}


DWORD CALLBACK ConInThread( HANDLE h )
{
   // starts a user input thread using the console
   //
   // takes console characters and sends them to the com port


   OVERLAPPED ov;
   HANDLE hconn = GetStdHandle(STD_INPUT_HANDLE);
   BOOL quit = FALSE;

   ZeroMemory(&ov,sizeof(ov));
   ov.hEvent = CreateEvent(NULL,FALSE,FALSE,NULL);
   if(ov.hEvent == INVALID_HANDLE_VALUE) {
      PrintError("E001_CreateEvent failed");
      SetCommMask(h,0);
      return 0;
   }

   char kb;
   INPUT_RECORD b1;

   SetConsoleMode(hconn,0);
   printf("press Esc or Ctrl+C to terminate\n");
    
   do {
      char buf[1000];
	  DWORD read = 0;

      SetConsoleMode(hconn,0);
   
      // wait for user to type something
      WaitForSingleObject(hconn,INFINITE);

      // read the console buffer
	  if(!ReadConsoleInput(hconn,&b1,1,&read)) {
         PrintError("E002_ReadConsoleInput failed...");
         quit  = TRUE;
      }
	  
	  kb=b1.Event.KeyEvent.uChar.AsciiChar;
	  //printf("read=%i, b1.keyevent: 0x%x ", read, kb);
      buf[0] = kb;
	  
      
	  if(read) {
         DWORD write=1;
         // check for Esc
		 // for debug purpose to find out the key value ...
		 // printf("Val 0x%x\n", kb);
			 
		 // terminate when Esc pressed
         if( buf[0] == 0x1b ||       // ESC
				// buf[0] == 0x11 ||       // Ctrl + Q
				// buf[0] == 0x18 ||       // Ctrl + X
				buf[0] == 0x3 ) {      // Ctrl + C
               quit = TRUE;
		       break;
         } else {
				 if (buf[0] != 0x0  && b1.Event.KeyEvent.bKeyDown) {
					// send it to the com port
					if( !WriteFile(h,buf,write,&write,&ov) ) {
						if(GetLastError() == ERROR_IO_PENDING) {
							if( !GetOverlappedResult(h,&ov,&write,TRUE) ) {
								PrintError("E003_GetOverlappedResult failed");
								quit = TRUE;
							}
						}
						// make the local output - echo is global var ...
						if (echo == 1) printf("%c", (unsigned char)buf[0]);
					} else {
						PrintError("E004_WriteFile failed");
						quit = TRUE;
					}
				} 
         }
      }

   } while(!quit);

   // tell terminal thread to quit
   if (!SetCommMask(h,0)) {
	   printf("SetCommMask-GetLastError: %i\n", GetLastError());
   }
   return 0;
}


void Terminal( HANDLE h )
{
   HANDLE hconn = GetStdHandle(STD_OUTPUT_HANDLE);
   DWORD mask;
   DWORD id, i;
   OVERLAPPED ov;


   FILE *stream;
   if (logo ==1) {
		stream = fopen( logname, "w" );
   }
   

   HANDLE hconin = CreateThread(NULL,0,ConInThread,h,0,&id);
   if( hconin == INVALID_HANDLE_VALUE ) {
      PrintError("E005_CreateThread failed");
      return;
   }
   CloseHandle(hconin);  // don't need this handle

   ZeroMemory(&ov,sizeof(ov));

   // create event for overlapped I/O
   ov.hEvent = CreateEvent(NULL,FALSE,FALSE,NULL);
   if(ov.hEvent == INVALID_HANDLE_VALUE)
      PrintError("E006_CreateEvent failed");

   // wait for received characters
   if(!SetCommMask(h,EV_RXCHAR))
      PrintError("E007_SetCommMask failed");

   while( 1 ) {
      // get the event mask
      if( !WaitCommEvent(h,&mask,&ov) ) {
         DWORD e = GetLastError();
         if( e == ERROR_IO_PENDING ) {
            DWORD r;
            if( !GetOverlappedResult(h,&ov,&r,TRUE) ) {
               PrintError("E008_GetOverlappedResult failed");
               break;
            }
         } else {
            PrintError("E009_WaitCommEvent failed");
            break;
         }
      }

      // if no event, then UI thread terminated with SetCommMask(h,0)
      if( mask == 0 ) {
         break;
      }

      if( mask & EV_RXCHAR) {
         char buf[1000];
         DWORD read;
         do {
            read = 0;
            if( !ReadFile(h,buf,sizeof(buf),&read,&ov) ) {
               if( GetLastError() == ERROR_IO_PENDING ) {
                  if( !GetOverlappedResult(h,&ov,&read,TRUE) ) {
                     PrintError("E010_GetOverlappedResult failed");
                     break;
                  }
               } else {
                  PrintError("E011_ReadFile failed");
                  break;
               }
            }
            if(read)
		 	  // WriteFile(hconn,buf,read,&read,NULL);
			  //printf("%s\n",buf);
			  for (i=0; i<read; i++) {
				  if (DisplayMode==0) printf("0x%x", (unsigned char)buf[i]);
				  if (DisplayMode==1) {
//***************** aqui comienza lo que agrego *************************
					  k=k+1;
					  //printf("itera %d %d",i,k);
					  if ((k-1)<170) {
						  bufo[k-1]=(unsigned char)buf[i];
						  //printf("%c", (unsigned char)buf[i]);
					  }
					  if ((k-1)==170) {
						  for (i1=0; i1<170; i1++) {
							  if (bufo[i1]=='$') {
								  aux1=i1;
								  for (j=0;j<5;j++) {
									  ident[j]=bufo[j+aux1+1];
								  }
								  ident[5]=0;
								  //printf("funciona\n");
								  //printf("%s",ident);
								  if (strcmp("GPGGA",ident)==0) {
									  contador=contador+1;
									  for (j=0;j<85;j++) {
										  datos_brutos[j]=bufo[j+aux1];
									  }
									  for (j=0;j<85;j++){   /*Busca las comas dentro de una linea*/
										  if (datos_brutos[j]==','){
											  cont1=cont1++;
											  comas[cont1-1]=j;
										  }
									  }
									  for (j=comas[1]+1;j<comas[2];j++){ 
										  latitud[j-comas[1]-1]=datos_brutos[j];
									  }
									  for (j=comas[2]+1;j<comas[3];j++){
										  latitud_dir=datos_brutos[j];
									  }
									  for (j=comas[3]+1;j<comas[4];j++){
										  longitud[j-comas[3]-1]=datos_brutos[j];
									  }
									  for (j=comas[4]+1;j<comas[5];j++){
										  longitud_dir=datos_brutos[j];
									  }
									  for (j=comas[8]+1;j<comas[9];j++){
										  altitud[j-comas[8]-1]=datos_brutos[j];
									  }
									  for (j=comas[0]+1;j<comas[1];j++){
										  hora[j-comas[0]-1]=datos_brutos[j]; 
									  }
									  for (j=0;j<11;j++){
										  if (j==0 || j==1) lat_hor[j]=latitud[j];
										  if (j==2 || j==3) lat_min_ent[j-2]=latitud[j];
										  if (j<11 && j>4) lat_min_dec[j-5]=latitud[j];
									  }
									  for (j=0;j<12;j++){
										  if (j==1 || j==2) long_hor[j-1]=longitud[j];
										  if (j==3 || j==4) long_min_ent[j-3]=longitud[j];
										  if (j>5 && j<12) long_min_dec[j-6]=longitud[j];
									  }
									  for (j=0;j<6;j++){
										  if (j==0 || j==1) hor[j]=hora[j];
										  if (j==2 || j==3) min[j-2]=hora[j];
										  if (j==4 || j==5) seg[j-4]=hora[j];
									  }  
									  a1=lat_hor[0]-'0';
									  a2=lat_hor[1]-'0';
									  lat_deg=10*a1+a2;
									  a3=lat_min_ent[0]-'0';
									  a4=lat_min_ent[1]-'0';
									  a5=lat_min_dec[0]-'0';
									  a6=lat_min_dec[1]-'0';
									  a7=lat_min_dec[2]-'0';
									  a8=lat_min_dec[3]-'0';
									  a9=lat_min_dec[4]-'0';
									  a10=lat_min_dec[5]-'0';
									  latmin=10*a3+a4+0.1*a5+0.01*a6+0.001*a7+0.0001*a8+0.00001*a9+0.000001*a10;
									  b1=long_hor[0]-'0';
									  b2=long_hor[1]-'0';
									  long_deg=10*b1+b2;
									  b3=long_min_ent[0]-'0';
									  b4=long_min_ent[1]-'0';
									  b5=long_min_dec[0]-'0';
									  b6=long_min_dec[1]-'0';
									  b7=long_min_dec[2]-'0';
									  b8=long_min_dec[3]-'0';
									  b9=long_min_dec[4]-'0';
									  b10=long_min_dec[5]-'0';
									  longmin=10*b3+b4+0.1*b5+0.01*b6+0.001*b7+0.0001*b8+0.00001*b9+0.000001*b10;
									  c1=altitud[0]-'0';
									  c2=altitud[1]-'0';
									  c3=altitud[2]-'0';
									  alt=100*c1+10*c2+c3;
									  d1=hor[0]-'0';
									  d2=hor[1]-'0';
									  hora1=10*d1+d2;
									  e1=min[0]-'0';
									  e2=min[1]-'0';
									  hora2=10*e1+e2;
									  f1=seg[0]-'0';
									  f2=seg[1]-'0';
									  hora3=10*f1+f2;

									  PHI=-(lat_deg+(latmin)/60)*3.1416/180;
									  THETA=-(long_deg+(longmin)/60)*3.1416/180;

									  /*Xmed=(radio+alt)*cos(PHI)*cos(THETA);
									  Ymed=(radio+alt)*cos(PHI)*sin(THETA);
									  Zmed=(radio+alt)*sin(PHI);*/
									  lambda=THETA-(long_central*3.1416/180);
									  eps=0.5*log((1+cos(PHI)*sin(lambda))/(1-cos(PHI)*sin(lambda)));
									  nu=atan(tan(PHI)/cos(lambda))-PHI;
									  ve=const_c*0.9996/(sqrt(1+er2*pow(cos(PHI),2)));
									  psi=er2*pow(eps,2)*pow(cos(PHI),2)/2;
									  AA1=sin(2*PHI);
									  AA2=AA1*pow(cos(PHI),2);
									  J2=PHI+AA1/2;
									  J4=(3*J2+AA2)/4;
									  J6=(5*J4+AA2*pow(cos(PHI),2))/3;
									  alpha=3*er2/4;
									  beta=5*pow(alpha,2)/3;
									  gamma=35*pow(alpha,3)/27;
									  BePHI=0.9996*const_c*(PHI-alpha*J2+beta*J4-gamma*J6);

									  xUTM=eps*ve*(1+psi/3)+500000;
									  yUTM=nu*ve*(1+psi)+BePHI+10000000;

									  if (contador==1) {
										  x[1]=xUTM;
										  x[0]=xUTM;
										  y[1]=yUTM;
										  y[0]=yUTM;
										  Xin=xUTM;
										  Yin=yUTM;
										  Xout=xUTM-Xin;
										  Yout=yUTM-Yin;
									  }									  
									  else {
										  Xout=xUTM-Xin;
										  Yout=yUTM-Yin;
										  x[0]=x[1];
										  x[1]=Xout;
										  desp_x=x[1]-x[0];
										  y[0]=y[1];
										  y[1]=Yout;
										  desp_y=y[1]-y[0];
										  aux2=pow(desp_x,2)+pow(desp_y,2);
										  desp=sqrt(aux2);
									  }
									  
									  //cargar variables
									  float alpha_e= 0.068067;

									  variables->desp_X=-sin(alpha_e)*desp_x+cos(alpha_e)*desp_y;
									  variables->desp_Y=-cos(alpha_e)*desp_x-sin(alpha_e)*desp_y;

                                      variables->ang_lat=lat_deg;
									  variables->min_lat=latmin;
									  variables->latitud=latitud_dir;

                                      variables->ang_long=long_deg;
									  variables->min_long=longmin;
									  variables->longitud=longitud_dir;

                                      // visualizar en pantalla

/*									  system("cls");
									  printf("Latitud: %d g %f m %c\n",lat_deg,latmin,latitud_dir);
									  printf("Longitud: %d g %f m %c\n",long_deg,longmin,longitud_dir);
									  printf("Altitud: %d\n",alt);
									  printf("Tiempo: %d h %d m %d s\n\n",hora1,hora2,hora3);
									  printf("UTM: (%f,%f)\n\n",xUTM,yUTM);
									  printf("Posicion: (%f,%f)\n",Xout,Yout);
									  printf("Desplazamiento en X:%f metros\n",desp_x);
									  printf("Desplazamiento en Y:%f metros\n",desp_y);
									  printf("Desplazamiento total:%f metros\n\n",aux2);

									  printf("%d iteraciones y %d lineas \n\n",i,contador);

*/
								  }
							  }
						  }
						  k=0;
					  }


//****************** hasta aqui llega mi intervencion ******************
				  }
				
				  if (DisplayMode==2) printf("%i", (unsigned char)buf[i]);
				  if (separator==0) printf(" ");
				  if (separator==1) printf("\n");
				  if (separator==2) printf("\t");
				  // separator==3 - do nothing!

				  if (logo ==1) {
		  			  // print to the logfile
					  if (DisplayMode==0) fprintf(stream,"0x%x", (unsigned char)buf[i]);
					  if (DisplayMode==1) fprintf(stream,"%c", (unsigned char)buf[i]);
					  if (DisplayMode==2) fprintf(stream,"%i", (unsigned char)buf[i]);
					  if (separator==0) fprintf(stream," ");
					  if (separator==1) fprintf(stream,"\n");
					  if (separator==2) fprintf(stream,"\t");
					  // separator==3 - do nothing!
				  }
			  }
		 } while(read);
      }
      mask = 0;
   }
   CloseHandle(ov.hEvent);  // close the event
   if (logo ==1) {
	 fclose( stream );
   }
  
  
 }

void wait4keypressed()
{
	  HANDLE hStdin; 
	  DWORD cNumRead, fdwMode, fdwSaveOldMode, i, quit; 
      INPUT_RECORD irInBuf[128]; 
 
      // Get the standard input handle. 
 
      hStdin = GetStdHandle(STD_INPUT_HANDLE); 
      if (hStdin == INVALID_HANDLE_VALUE) 
        printf("Problem:GetStdHandle"); 
 
      // Save the current input mode, to be restored on exit. 
      if (! GetConsoleMode(hStdin, &fdwSaveOldMode) ) 
        printf("Problem:GetConsoleMode"); 
 
       //Enable the window and mouse input events. 
 
      fdwMode = ENABLE_WINDOW_INPUT | ENABLE_MOUSE_INPUT; 
      if (! SetConsoleMode(hStdin, fdwMode) ) 
        printf("Problem:SetConsoleMode"); 
 
      // Loop to read and handle the input events. 
	  quit = 1;

      while (quit) 
	  { 
        // Wait for the events. 
        if (! ReadConsoleInput( 
                hStdin,      // input buffer handle 
                irInBuf,     // buffer to read into 
                128,         // size of read buffer 
                &cNumRead) ) // number of records read 
            printf("\nProblem:ReadConsoleInput\n"); 
 
        // check if key pressed ...
        for (i = 0; i < cNumRead; i++) 
        {
			if (irInBuf[i].EventType == KEY_EVENT) quit=0;
        }
	  } 

	  // rest the console mode
	  SetConsoleMode(hStdin, fdwSaveOldMode);
}



int main_gps()
{

int inicio=2;

   // copy the name of the executable into a global variable
//   strcpy(exename,argv[0]);
   strcpy(exename,"GPS");

   if(inicio>1) {
      // open port for overlapped I/O
      HANDLE h = CreateFile(puerto,
                    GENERIC_READ|GENERIC_WRITE,
                    0,NULL,
                    OPEN_EXISTING,FILE_FLAG_OVERLAPPED,NULL);

      if(h == INVALID_HANDLE_VALUE) {
         PrintError("E012_Failed to open port");
      } else {
         // set timeouts
         COMMTIMEOUTS cto = { 2, 1, 1, 0, 0 };
         DCB dcb;
         if(!SetCommTimeouts(h,&cto))
            PrintError("E013_SetCommTimeouts failed");

         // set DCB
         memset(&dcb,0,sizeof(dcb));
         dcb.DCBlength = sizeof(dcb);
         dcb.BaudRate = 19200;
/*         if(argc>2 && atoi(argv[2]) )
            dcb.BaudRate = atoi(argv[2]);*/
         DisplayMode = 1;
/*		 if(argc>3) {
			if (argv[3][0] == 'h') DisplayMode = 0;
			if (argv[3][0] == 'a') DisplayMode = 1;
			if (argv[3][0] == 'd') DisplayMode = 2;
		 }*/
         separator = 3;
/*		 if(argc>4) {
			if (argv[4][0] == 's') separator = 0;
			if (argv[4][0] == 'n') separator = 1;		 
			if (argv[4][0] == 't') separator = 2;		 
			if (argv[4][0] == 'e') separator = 3;		 
		 }*/
         echo = 0;
/*		 if(argc>5) {
			if (argv[5][0] == 'n') echo = 0;
			if (argv[5][0] == 'y') echo = 1;
		 }*/
		logo = 0;
/*		 if(argc>6) {
			logo =1;
			strcpy(logname, "datos.txt");
		 }*/


         dcb.fBinary = 1;
		 dcb.fDtrControl = DTR_CONTROL_ENABLE;
		 dcb.fRtsControl = RTS_CONTROL_ENABLE;
		 // dcb.fOutxCtsFlow = 1;
  		 // dcb.fRtsControl = DTR_CONTROL_HANDSHAKE;

         dcb.Parity = NOPARITY;
         dcb.StopBits = ONESTOPBIT;
         dcb.ByteSize = 8;

         if(!SetCommState(h,&dcb))
            PrintError("E014_SetCommState failed");

		

         // start terminal
         Terminal(h);
	     CloseHandle(h);
	
      }
   } else {
	  printf("Commandline Serial Terminal - Version 1.1 by A. Schmidt, Oct 2001\n");
	  printf("Lancaster University - http://www.comp.lancs.ac.uk/~albrecht/\n\n");
	  help();
	  wait4keypressed();
	}
   return 0;
}
//"com1 57600 hex space"
/*int main(int argc, char **argv){
   return main_gps();
}
*/